package dev.fyloz.colorrecipesexplorer.config.security

import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.emergencyMode
import dev.fyloz.colorrecipesexplorer.logic.users.JwtLogic
import dev.fyloz.colorrecipesexplorer.logic.users.UserDetailsLogic
import dev.fyloz.colorrecipesexplorer.logic.users.UserLogic
import dev.fyloz.colorrecipesexplorer.model.account.Permission
import dev.fyloz.colorrecipesexplorer.model.account.User
import mu.KotlinLogging
import org.slf4j.Logger
import org.springframework.boot.context.properties.EnableConfigurationProperties
import org.springframework.context.annotation.Bean
import org.springframework.context.annotation.Configuration
import org.springframework.context.annotation.Lazy
import org.springframework.context.annotation.Profile
import org.springframework.core.env.Environment
import org.springframework.http.HttpMethod
import org.springframework.security.config.annotation.authentication.builders.AuthenticationManagerBuilder
import org.springframework.security.config.annotation.method.configuration.EnableGlobalMethodSecurity
import org.springframework.security.config.annotation.web.builders.HttpSecurity
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity
import org.springframework.security.config.annotation.web.configuration.WebSecurityConfigurerAdapter
import org.springframework.security.config.http.SessionCreationPolicy
import org.springframework.security.core.AuthenticationException
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder
import org.springframework.security.web.AuthenticationEntryPoint
import org.springframework.stereotype.Component
import org.springframework.web.cors.CorsConfiguration
import org.springframework.web.cors.UrlBasedCorsConfigurationSource
import javax.annotation.PostConstruct
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse

private const val angularDevServerOrigin = "http://localhost:4200"
private const val rootUserFirstName = "Root"
private const val rootUserLastName = "User"

abstract class BaseSecurityConfig(
    private val userDetailsLogic: UserDetailsLogic,
    private val jwtLogic: JwtLogic,
    private val environment: Environment,
    protected val securityProperties: CreSecurityProperties
) : WebSecurityConfigurerAdapter() {
    protected abstract val logger: Logger

    protected val passwordEncoder = BCryptPasswordEncoder()
    var debugMode = false

    @Bean
    open fun passwordEncoder() =
        passwordEncoder

    @Bean
    open fun corsConfigurationSource() =
        UrlBasedCorsConfigurationSource().apply {
            registerCorsConfiguration("/**", CorsConfiguration().apply {
                allowedOrigins = listOf(angularDevServerOrigin)
                allowedMethods = listOf(
                    HttpMethod.GET.name,
                    HttpMethod.POST.name,
                    HttpMethod.PUT.name,
                    HttpMethod.DELETE.name,
                    HttpMethod.OPTIONS.name,
                    HttpMethod.HEAD.name
                )
                allowCredentials = true
            }.applyPermitDefaultValues())
        }

    override fun configure(authBuilder: AuthenticationManagerBuilder) {
        authBuilder.userDetailsService(userDetailsLogic).passwordEncoder(passwordEncoder)
    }

    override fun configure(http: HttpSecurity) {
        http
            .headers().frameOptions().disable()
            .and()
            .csrf().disable()
            .addFilter(
                JwtAuthenticationFilter(
                    authenticationManager(),
                    jwtLogic,
                    securityProperties,
                    this::updateUserLoginTime
                )
            )
            .addFilter(
                JwtAuthorizationFilter(jwtLogic, authenticationManager(), userDetailsLogic)
            )
            .sessionManagement().sessionCreationPolicy(SessionCreationPolicy.STATELESS)
            .and()
            .authorizeRequests()
            .antMatchers("/api/config/**").permitAll()    // Allow access to logo and icon
            .antMatchers("/api/login").permitAll()  // Allow access to login
            .antMatchers("**").fullyAuthenticated()

        if (debugMode) {
            http
                .cors()
        }
    }

    @PostConstruct
    fun initDebugMode() {
        debugMode = "debug" in environment.activeProfiles
        if (debugMode) logger.warn("Debug mode is enabled, security will be decreased!")
    }

    protected open fun updateUserLoginTime(userId: Long) {
    }
}

@Configuration
@Profile("!emergency")
@EnableWebSecurity
@EnableGlobalMethodSecurity(prePostEnabled = true)
@EnableConfigurationProperties(CreSecurityProperties::class)
class SecurityConfig(
    @Lazy userDetailsLogic: UserDetailsLogic,
    @Lazy private val userLogic: UserLogic,
    jwtLogic: JwtLogic,
    environment: Environment,
    securityProperties: CreSecurityProperties
) : BaseSecurityConfig(userDetailsLogic, jwtLogic, environment, securityProperties) {
    override val logger = KotlinLogging.logger {}

    @PostConstruct
    fun initWebSecurity() {
        if (emergencyMode) {
            logger.error("Emergency mode is enabled, system users will not be created")
            return
        }

        createRootUser()
    }

    override fun updateUserLoginTime(userId: Long) {
        userLogic.updateLastLoginTime(userId)
    }

    private fun createRootUser() {
        if (securityProperties.root == null) {
            throw InvalidSystemUserException("root", "cre.security.root configuration is not defined")
        }

        with(securityProperties.root!!) {
            if (!userLogic.existsById(this.id)) {
                userLogic.save(
                    User(
                        id = this.id,
                        firstName = rootUserFirstName,
                        lastName = rootUserLastName,
                        password = passwordEncoder.encode(this.password),
                        isSystemUser = true,
                        permissions = mutableSetOf(Permission.ADMIN)
                    )
                )
            }
        }
    }
}

@Configuration
@Profile("emergency")
@EnableConfigurationProperties(CreSecurityProperties::class)
class EmergencySecurityConfig(
    userDetailsLogic: UserDetailsLogic,
    jwtLogic: JwtLogic,
    environment: Environment,
    securityProperties: CreSecurityProperties
) : BaseSecurityConfig(userDetailsLogic, jwtLogic, environment, securityProperties) {
    override val logger = KotlinLogging.logger {}

    init {
        emergencyMode = true
    }
}

@Component
class RestAuthenticationEntryPoint : AuthenticationEntryPoint {
    override fun commence(
        request: HttpServletRequest,
        response: HttpServletResponse,
        authException: AuthenticationException
    ) = response.sendError(HttpServletResponse.SC_UNAUTHORIZED, "Unauthorized")
}

private class InvalidSystemUserException(userType: String, message: String) :
    RuntimeException("Invalid $userType user: $message")
