package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.config.annotations.RequireDatabase
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.MixTypeRepository
import org.springframework.context.annotation.Lazy
import org.springframework.stereotype.Service

interface MixTypeLogic : NamedModelService<MixType, MixTypeRepository> {
    /** Checks if a [MixType] with the given [name] and [materialType] exists. */
    fun existsByNameAndMaterialType(name: String, materialType: MaterialType): Boolean

    /** Gets the mix type with the given [material]. */
    fun getByMaterial(material: Material): MixType

    /** Gets the [MixType] with the given [name] and [materialType]. */
    fun getByNameAndMaterialType(name: String, materialType: MaterialType): MixType

    /** Returns a [MixType] for the given [name] and [materialType]. If a mix type with these does not already exists, it will be created. */
    fun getOrCreateForNameAndMaterialType(name: String, materialType: MaterialType): MixType

    /** Returns a new and persisted [MixType] with the given [name] and [materialType]. */
    fun saveForNameAndMaterialType(name: String, materialType: MaterialType): MixType

    /** Returns the given [mixType] updated with the given [name] and [materialType]. */
    fun updateForNameAndMaterialType(mixType: MixType, name: String, materialType: MaterialType): MixType
}

@Service
@RequireDatabase
class DefaultMixTypeLogic(
    mixTypeRepository: MixTypeRepository,
    @Lazy val materialLogic: MaterialLogic
) :
    AbstractNamedModelService<MixType, MixTypeRepository>(mixTypeRepository), MixTypeLogic {
    override fun idNotFoundException(id: Long) = mixTypeIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = mixTypeIdAlreadyExistsException(id)
    override fun nameNotFoundException(name: String) = mixTypeNameNotFoundException(name)
    override fun nameAlreadyExistsException(name: String) = mixTypeNameAlreadyExistsException(name)

    override fun existsByNameAndMaterialType(name: String, materialType: MaterialType): Boolean =
        repository.existsByNameAndMaterialType(name, materialType)

    override fun getByMaterial(material: Material): MixType =
        repository.findByMaterial(material) ?: throw nameNotFoundException(material.name)

    override fun getByNameAndMaterialType(name: String, materialType: MaterialType): MixType =
        repository.findByNameAndMaterialType(name, materialType)
            ?: throw MixTypeNameAndMaterialTypeNotFoundException(name, materialType)

    override fun getOrCreateForNameAndMaterialType(name: String, materialType: MaterialType): MixType =
        if (existsByNameAndMaterialType(name, materialType))
            getByNameAndMaterialType(name, materialType)
        else
            saveForNameAndMaterialType(name, materialType)

    override fun save(entity: MixType): MixType {
        if (materialLogic.existsByName(entity.name))
            throw materialNameAlreadyExistsException(entity.name)
        return super.save(entity)
    }

    override fun saveForNameAndMaterialType(name: String, materialType: MaterialType): MixType =
        save(
            mixType(
                name = name,
                material = material(
                    name = name,
                    inventoryQuantity = Float.MIN_VALUE,
                    isMixType = true,
                    materialType = materialType
                )
            )
        )

    override fun updateForNameAndMaterialType(mixType: MixType, name: String, materialType: MaterialType): MixType =
        update(mixType.copy(material = mixType.material.copy(name = name, materialType = materialType)))

    override fun delete(entity: MixType) {
        if (!repository.canBeDeleted(entity.id!!)) throw cannotDeleteMixTypeException(entity)
        super.delete(entity)
    }
}
