package dev.fyloz.colorrecipesexplorer.logic.users

import dev.fyloz.colorrecipesexplorer.SpringUserDetails
import dev.fyloz.colorrecipesexplorer.SpringUserDetailsService
import dev.fyloz.colorrecipesexplorer.config.annotations.RequireDatabase
import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.account.Permission
import dev.fyloz.colorrecipesexplorer.model.account.User
import dev.fyloz.colorrecipesexplorer.model.account.UserDetails
import dev.fyloz.colorrecipesexplorer.model.account.user
import org.springframework.context.annotation.Profile
import org.springframework.security.core.userdetails.UsernameNotFoundException
import org.springframework.stereotype.Service

interface UserDetailsLogic : SpringUserDetailsService {
    /** Loads an [User] for the given [id]. */
    fun loadUserById(id: Long, ignoreDefaultGroupUsers: Boolean = false): UserDetails
}

@Service
@RequireDatabase
class DefaultUserDetailsLogic(
    private val userLogic: UserLogic
) : UserDetailsLogic {
    override fun loadUserByUsername(username: String): UserDetails {
        try {
            return loadUserById(username.toLong(), true)
        } catch (ex: NotFoundException) {
            throw UsernameNotFoundException(username)
        }
    }

    override fun loadUserById(id: Long, ignoreDefaultGroupUsers: Boolean): UserDetails {
        val user = userLogic.getById(
            id,
            ignoreDefaultGroupUsers = ignoreDefaultGroupUsers,
            ignoreSystemUsers = false
        )
        return UserDetails(user)
    }
}

@Service
@Profile("emergency")
class EmergencyUserDetailsLogic(
    securityProperties: CreSecurityProperties
) : UserDetailsLogic {
    private val users: Set<User>

    init {
        if (securityProperties.root == null) {
            throw NullPointerException("The root user has not been configured")
        }

        users = setOf(
            // Add root user
            with(securityProperties.root!!) {
                user(
                    id = this.id,
                    plainPassword = this.password,
                    permissions = mutableSetOf(Permission.ADMIN)
                )
            }
        )
    }

    override fun loadUserByUsername(username: String): SpringUserDetails {
        return loadUserById(username.toLong(), true)
    }

    override fun loadUserById(id: Long, ignoreDefaultGroupUsers: Boolean): UserDetails {
        val user = users.firstOrNull { it.id == id }
            ?: throw UsernameNotFoundException(id.toString())

        return UserDetails(user)
    }
}
