package dev.fyloz.colorrecipesexplorer.model

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.dtos.MaterialDto
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.CannotDeleteException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import javax.persistence.*
import javax.validation.constraints.Min

const val SIMDUT_FILES_PATH = "pdf/simdut"

@Entity
@Table(name = "material")
data class Material(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    override val id: Long?,

    @Column(unique = true)
    val name: String,

    @Column(name = "inventory_quantity")
    val inventoryQuantity: Float,

    @Column(name = "mix_type")
    val isMixType: Boolean,

    @ManyToOne
    @JoinColumn(name = "material_type_id")
    val materialType: MaterialType?
) : ModelEntity {
    companion object {
        fun getSimdutFilePath(name: String) =
            "${Constants.FilePaths.simdut}/$name.pdf"
    }
}

data class MaterialQuantityDto(
    val material: Long,

    @field:Min(0, message = VALIDATION_SIZE_GE_ZERO)
    val quantity: Float
)

// === DSL ===

fun material(
    id: Long? = null,
    name: String = "name",
    inventoryQuantity: Float = 0f,
    isMixType: Boolean = false,
    materialType: MaterialType? = materialType(),
    op: Material.() -> Unit = {}
) = Material(id, name, inventoryQuantity, isMixType, materialType).apply(op)

fun material(
    material: Material,
    id: Long? = null,
    name: String? = null,
) = Material(
    id ?: material.id, name
        ?: material.name, material.inventoryQuantity, material.isMixType, material.materialType
)

@Deprecated("Temporary DSL for transition")
fun material(
    dto: MaterialDto
) = Material(dto.id, dto.name, dto.inventoryQuantity, dto.isMixType, dto.materialType)

@Deprecated("Temporary DSL for transition")
fun materialDto(
    entity: Material
) = MaterialDto(entity.id!!, entity.name, entity.inventoryQuantity, entity.isMixType, entity.materialType!!)

fun materialQuantityDto(
    materialId: Long,
    quantity: Float,
    op: MaterialQuantityDto.() -> Unit = {}
) = MaterialQuantityDto(materialId, quantity).apply(op)

// ==== Exceptions ====
private const
val MATERIAL_NOT_FOUND_EXCEPTION_TITLE = "Material not found"
private const val MATERIAL_ALREADY_EXISTS_EXCEPTION_TITLE = "Material already exists"
private const val MATERIAL_CANNOT_DELETE_EXCEPTION_TITLE = "Cannot delete material"
private const val MATERIAL_EXCEPTION_ERROR_CODE = "material"

fun materialIdNotFoundException(id: Long) =
    NotFoundException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_NOT_FOUND_EXCEPTION_TITLE,
        "A material with the id $id could not be found",
        id
    )

fun materialNameNotFoundException(name: String) =
    NotFoundException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_NOT_FOUND_EXCEPTION_TITLE,
        "A material with the name $name could not be found",
        name,
        "name"
    )

fun materialIdAlreadyExistsException(id: Long) =
    AlreadyExistsException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_ALREADY_EXISTS_EXCEPTION_TITLE,
        "A material with the id $id already exists",
        id
    )

fun materialNameAlreadyExistsException(name: String) =
    AlreadyExistsException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_ALREADY_EXISTS_EXCEPTION_TITLE,
        "A material with the name $name already exists",
        name,
        "name"
    )

fun cannotDeleteMaterialException(material: Material) =
    CannotDeleteException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_CANNOT_DELETE_EXCEPTION_TITLE,
        "Cannot delete the material ${material.name} because one or more recipes depends on it"
    )
