package dev.fyloz.colorrecipesexplorer.logic

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.config.security.defaultGroupCookieName
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.account.*
import dev.fyloz.colorrecipesexplorer.repository.GroupRepository
import dev.fyloz.colorrecipesexplorer.repository.UserRepository
import dev.fyloz.colorrecipesexplorer.logic.users.*
import org.junit.jupiter.api.*
import org.springframework.mock.web.MockHttpServletResponse
import org.springframework.security.core.userdetails.UsernameNotFoundException
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder
import java.util.*
import javax.servlet.http.Cookie
import javax.servlet.http.HttpServletRequest
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertNotNull
import kotlin.test.assertTrue

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class UserLogicTest :
  AbstractExternalModelServiceTest<User, UserSaveDto, UserUpdateDto, UserLogic, UserRepository>() {
    private val passwordEncoder = BCryptPasswordEncoder()

    override val entity: User = user(id = 0L, passwordEncoder = passwordEncoder)
    override val anotherEntity: User = user(id = 1L, passwordEncoder = passwordEncoder)
    private val entityDefaultGroupUser = user(id = 2L, isDefaultGroupUser = true, passwordEncoder = passwordEncoder)
    private val entitySystemUser = user(id = 3L, isSystemUser = true, passwordEncoder = passwordEncoder)
    private val group = group(id = 0L)
    override val entitySaveDto: UserSaveDto = spy(userSaveDto(passwordEncoder, id = 0L))
    override val entityUpdateDto: UserUpdateDto = spy(userUpdateDto(id = 0L))

    override val repository: UserRepository = mock()
    private val groupService: GroupLogic = mock()
    override val logic: UserLogic = spy(DefaultUserLogic(repository, groupService))

    private val entitySaveDtoUser = User(
      entitySaveDto.id,
      entitySaveDto.firstName,
      entitySaveDto.lastName,
      passwordEncoder.encode(entitySaveDto.password),
      isDefaultGroupUser = false,
      isSystemUser = false,
      group = null,
      permissions = entitySaveDto.permissions
    )

    @AfterEach
    override fun afterEach() {
        reset(groupService)
        super.afterEach()
    }

    // existsByFirstNameAndLastName()

    @Test
    fun `existsByFirstNameAndLastName() returns true when an user with the given first name and last name exists`() {
        whenever(repository.existsByFirstNameAndLastName(entity.firstName, entity.lastName)).doReturn(true)

        val found = logic.existsByFirstNameAndLastName(entity.firstName, entity.lastName)

        assertTrue(found)
    }

    @Test
    fun `existsByFirstNameAndLastName() returns false when no user with the given first name and last name exists`() {
        whenever(repository.existsByFirstNameAndLastName(entity.firstName, entity.lastName)).doReturn(false)

        val found = logic.existsByFirstNameAndLastName(entity.firstName, entity.lastName)

        assertFalse(found)
    }

    // getById()

    @Test
    fun `getById() throws NotFoundException when the corresponding user is a default group user`() {
        whenever(repository.findById(entityDefaultGroupUser.id)).doReturn(Optional.of(entityDefaultGroupUser))

        assertThrows<NotFoundException> {
            logic.getById(
              entityDefaultGroupUser.id,
              ignoreDefaultGroupUsers = true,
              ignoreSystemUsers = false
            )
        }.assertErrorCode()
    }

    @Test
    fun `getById() throws NotFoundException when the corresponding user is a system user`() {
        whenever(repository.findById(entitySystemUser.id)).doReturn(Optional.of(entitySystemUser))

        assertThrows<NotFoundException> {
            logic.getById(
              entitySystemUser.id,
              ignoreDefaultGroupUsers = false,
              ignoreSystemUsers = true
            )
        }.assertErrorCode()
    }

    // getByGroup()

    @Test
    fun `getByGroup() returns all the users with the given group from the repository`() {
        whenever(repository.findAllByGroup(group)).doReturn(entityList)

        val found = logic.getByGroup(group)

        assertTrue(found.containsAll(entityList))
        assertTrue(entityList.containsAll(found))
    }

    @Test
    fun `getByGroup() returns an empty list when there is no user with the given group in the repository`() {
        whenever(repository.findAllByGroup(group)).doReturn(listOf())

        val found = logic.getByGroup(group)

        assertTrue(found.isEmpty())
    }

    // getDefaultGroupUser()

    @Test
    fun `getDefaultGroupUser() returns the default user of the given group from the repository`() {
        whenever(repository.findByIsDefaultGroupUserIsTrueAndGroupIs(group)).doReturn(entityDefaultGroupUser)

        val found = logic.getDefaultGroupUser(group)

        assertEquals(entityDefaultGroupUser, found)
    }

    // save()

    override fun `save() saves in the repository and returns the saved value`() {
        whenever(repository.save(entity)).doReturn(entity)
        doReturn(false).whenever(repository).existsByFirstNameAndLastName(entity.firstName, entity.lastName)

        val found = logic.save(entity)

        verify(repository).save(entity)
        assertEquals(entity, found)
    }

    @Test
    fun `save() throws AlreadyExistsException when firstName and lastName exists`() {
        doReturn(true).whenever(repository).existsByFirstNameAndLastName(entity.firstName, entity.lastName)

        assertThrows<AlreadyExistsException> { logic.save(entity) }
          .assertErrorCode("fullName")
    }

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        withBaseSaveDtoTest(entity, entitySaveDto, logic, {
            argThat {
                this.id == entity.id && this.firstName == entity.firstName && this.lastName == entity.lastName
            }
        })
    }

    @Test
    fun `save(dto) calls and returns save() with the created user`() {
        doReturn(entitySaveDtoUser).whenever(logic).save(any<User>())

        val found = logic.save(entitySaveDto)

        verify(logic).save(argThat<User> { this.id == entity.id && this.firstName == entity.firstName && this.lastName == entity.lastName })
        assertEquals(entitySaveDtoUser, found)
    }

    // update()

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() =
      withBaseUpdateDtoTest(entity, entityUpdateDto, logic, { any() })

    @Test
    fun `update() throws AlreadyExistsException when a different user with the given first name and last name exists`() {
        whenever(repository.findByFirstNameAndLastName(entity.firstName, entity.lastName)).doReturn(
          entityDefaultGroupUser
        )
        doReturn(entity).whenever(logic).getById(eq(entity.id), any(), any())

        assertThrows<AlreadyExistsException> {
            logic.update(
              entity,
              true,
              ignoreSystemUsers = true
            )
        }.assertErrorCode("fullName")
    }
}

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class GroupLogicTest :
  AbstractExternalNamedModelServiceTest<Group, GroupSaveDto, GroupUpdateDto, GroupLogic, GroupRepository>() {
    private val userService: UserLogic = mock()
    override val repository: GroupRepository = mock()
    override val logic: DefaultGroupLogic = spy(DefaultGroupLogic(userService, repository))

    override val entity: Group = group(id = 0L, name = "group")
    override val anotherEntity: Group = group(id = 1L, name = "another group")
    override val entitySaveDto: GroupSaveDto = spy(groupSaveDto(name = "group"))
    override val entityUpdateDto: GroupUpdateDto = spy(groupUpdateDto(id = 0L, name = "group"))
    override val entityWithEntityName: Group = group(id = 2L, name = entity.name)

    private val groupUserId = 1000000L + entity.id!!
    private val groupUser = user(passwordEncoder = BCryptPasswordEncoder(), id = groupUserId, group = entity)

    @BeforeEach
    override fun afterEach() {
        reset(userService)
        super.afterEach()
    }

    // getUsersForGroup()

    @Test
    fun `getUsersForGroup() returns all users in the given group`() {
        val group = group(id = 1L)

        doReturn(group).whenever(logic).getById(group.id!!)
        whenever(userService.getByGroup(group)).doReturn(listOf(groupUser))

        val found = logic.getUsersForGroup(group.id!!)

        assertTrue(found.contains(groupUser))
        assertTrue(found.size == 1)
    }

    @Test
    fun `getUsersForGroup() returns empty collection when the given group contains any user`() {
        doReturn(entity).whenever(logic).getById(entity.id!!)

        val found = logic.getUsersForGroup(entity.id!!)

        assertTrue(found.isEmpty())
    }

    // getRequestDefaultGroup()

    @Test
    fun `getRequestDefaultGroup() returns the group contained in the cookie of the HTTP request`() {
        val cookies: Array<Cookie> = arrayOf(Cookie(defaultGroupCookieName, groupUserId.toString()))
        val request: HttpServletRequest = mock()

        whenever(request.cookies).doReturn(cookies)
        whenever(userService.getById(eq(groupUserId), any(), any())).doReturn(groupUser)

        val found = logic.getRequestDefaultGroup(request)

        assertEquals(entity, found)
    }

    @Test
    fun `getRequestDefaultGroup() throws NoDefaultGroupException when the HTTP request does not contains a cookie for the default group`() {
        val request: HttpServletRequest = mock()

        whenever(request.cookies).doReturn(arrayOf())

        assertThrows<NoDefaultGroupException> { logic.getRequestDefaultGroup(request) }
    }

    // setResponseDefaultGroup()

    @Test
    fun `setResponseDefaultGroup() the default group cookie has been added to the given HTTP response with the given group id`() {
        val response = MockHttpServletResponse()

        whenever(userService.getDefaultGroupUser(entity)).doReturn(groupUser)
        doReturn(entity).whenever(logic).getById(entity.id!!)

        logic.setResponseDefaultGroup(entity.id!!, response)
        val found = response.getCookie(defaultGroupCookieName)

        assertNotNull(found)
        assertEquals(defaultGroupCookieName, found.name)
        assertEquals(groupUserId.toString(), found.value)
        assertEquals(defaultGroupCookieMaxAge, found.maxAge)
        assertTrue(found.isHttpOnly)
        assertTrue(found.secure)
    }

    // save()

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        withBaseSaveDtoTest(entity, entitySaveDto, logic)
    }

    // update()

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() =
      withBaseUpdateDtoTest(entity, entityUpdateDto, logic, { any() })
}

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class UserUserDetailsLogicTest {
    private val userLogic: UserLogic = mock()
    private val logic = spy(DefaultUserDetailsLogic(userLogic))

    private val user = user(id = 0L)

    @BeforeEach
    fun beforeEach() {
        reset(userLogic, logic)
    }

    // loadUserByUsername()

    @Test
    fun `loadUserByUsername() calls loadUserByUserId() with the given username as an id`() {
        whenever(userLogic.getById(eq(user.id), any(), any())).doReturn(user)
        doReturn(UserDetails(user(id = user.id, plainPassword = user.password)))
            .whenever(logic).loadUserById(user.id)

        logic.loadUserByUsername(user.id.toString())

        verify(logic).loadUserById(eq(user.id), any())
    }

    @Test
    fun `loadUserByUsername() throws UsernameNotFoundException when no user with the given id exists`() {
        whenever(userLogic.getById(eq(user.id), any(), any())).doThrow(
          userIdNotFoundException(user.id)
        )

        assertThrows<UsernameNotFoundException> { logic.loadUserByUsername(user.id.toString()) }
    }

    // loadUserByUserId

    @Test
    fun `loadUserByUserId() returns an User corresponding to the user with the given id`() {
        whenever(userLogic.getById(eq(user.id), any(), any())).doReturn(user)

        val found = logic.loadUserById(user.id)

        assertEquals(user.id, found.username.toLong())
        assertEquals(user.password, found.password)
    }
}
