package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.dtos.RecipeStepDto
import dev.fyloz.colorrecipesexplorer.model.RecipeGroupInformation
import dev.fyloz.colorrecipesexplorer.model.RecipeStep
import dev.fyloz.colorrecipesexplorer.model.account.Group
import dev.fyloz.colorrecipesexplorer.service.RecipeStepService
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertDoesNotThrow
import org.junit.jupiter.api.assertThrows
import kotlin.test.assertTrue

class DefaultRecipeStepLogicTest {
    private val recipeStepServiceMock = mockk<RecipeStepService>()

    private val recipeStepLogic = spyk(DefaultRecipeStepLogic(recipeStepServiceMock))

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    @Test
    fun validateGroupInformationSteps_normalBehavior_callsValidateSteps() {
        // Arrange
        every { recipeStepLogic.validateSteps(any()) } just runs

        val group = Group(1L, "Unit test group")
        val steps = mutableSetOf(RecipeStep(1L, 1, "A message"))
        val groupInfo = RecipeGroupInformation(1L, group, "A note", steps)

        // Act
        recipeStepLogic.validateGroupInformationSteps(groupInfo)

        // Assert
        verify {
            recipeStepLogic.validateSteps(any()) // TODO replace with actual steps dtos when RecipeGroupInformation updated
        }
    }

    @Test
    fun validateGroupInformationSteps_stepSetIsNull_doesNothing() {
        // Arrange
        every { recipeStepLogic.validateSteps(any()) } just runs

        val group = Group(1L, "Unit test group")
        val groupInfo = RecipeGroupInformation(1L, group, "A note", null)

        // Act
        recipeStepLogic.validateGroupInformationSteps(groupInfo)

        // Assert
        verify(exactly = 0) {
            recipeStepLogic.validateSteps(any()) // TODO replace with actual steps dtos when RecipeGroupInformation updated
        }
    }

    @Test
    fun validateGroupInformationSteps_invalidSteps_throwsInvalidGroupStepsPositionsException() {
        // Arrange
        val errors = setOf(InvalidStepsPositionsError("error", "An unit test error"))
        every { recipeStepLogic.validateSteps(any()) } throws InvalidStepsPositionsException(errors)

        val group = Group(1L, "Unit test group")
        val steps = mutableSetOf(RecipeStep(1L, 1, "A message"))
        val groupInfo = RecipeGroupInformation(1L, group, "A note", steps)

        // Act
        // Assert
        assertThrows<InvalidGroupStepsPositionsException> { recipeStepLogic.validateGroupInformationSteps(groupInfo) }
    }

    @Test
    fun validateSteps_normalBehavior_doesNothing() {
        // Arrange
        val recipeSteps = setOf(
            RecipeStepDto(1L, 1, "A message"),
            RecipeStepDto(2L, 2, "Another message")
        )

        // Act
        // Assert
        assertDoesNotThrow { recipeStepLogic.validateSteps(recipeSteps) }
    }

    @Test
    fun validateSteps_emptyStepSet_doesNothing() {
        // Arrange
        val recipeSteps = setOf<RecipeStepDto>()

        // Act
        // Assert
        assertDoesNotThrow { recipeStepLogic.validateSteps(recipeSteps) }
    }

    @Test
    fun validateSteps_hasInvalidPositions_throwsInvalidStepsPositionsException() {
        // Arrange
        val recipeSteps = setOf(
            RecipeStepDto(1L, 2, "A message"),
            RecipeStepDto(2L, 3, "Another message")
        )

        // Act
        // Assert
        assertThrows<InvalidStepsPositionsException> { recipeStepLogic.validateSteps(recipeSteps) }
    }

    @Test
    fun validateSteps_firstStepPositionInvalid_returnsInvalidStepValidationError() {
        // Arrange
        val recipeSteps = setOf(
            RecipeStepDto(1L, 2, "A message"),
            RecipeStepDto(2L, 3, "Another message")
        )

        // Act
        val exception = assertThrows<InvalidStepsPositionsException> { recipeStepLogic.validateSteps(recipeSteps) }

        // Assert
        assertTrue {
            exception.errors.any { it.type == RecipeStepDto.VALIDATION_ERROR_CODE_INVALID_FIRST_STEP }
        }
    }

    @Test
    fun validateSteps_duplicatedPositions_returnsInvalidStepValidationError() {
        // Arrange
        val recipeSteps = setOf(
            RecipeStepDto(1L, 1, "A message"),
            RecipeStepDto(2L, 1, "Another message")
        )

        // Act
        val exception = assertThrows<InvalidStepsPositionsException> { recipeStepLogic.validateSteps(recipeSteps) }

        // Assert
        assertTrue {
            exception.errors.any { it.type == RecipeStepDto.VALIDATION_ERROR_CODE_DUPLICATED_STEPS_POSITION }
        }
    }

    @Test
    fun validateSteps_gapsInPositions_returnsInvalidStepValidationError() {
        // Arrange
        val recipeSteps = setOf(
            RecipeStepDto(1L, 1, "A message"),
            RecipeStepDto(2L, 3, "Another message")
        )

        // Act
        val exception = assertThrows<InvalidStepsPositionsException> { recipeStepLogic.validateSteps(recipeSteps) }

        // Assert
        assertTrue {
            exception.errors.any { it.type == RecipeStepDto.VALIDATION_ERROR_CODE_GAP_BETWEEN_STEPS_POSITIONS }
        }
    }
}

//@TestInstance(TestInstance.Lifecycle.PER_CLASS)
//class RecipeStepLogicTest :
//  AbstractModelServiceTest<RecipeStep, RecipeStepLogic, RecipeStepRepository>() {
//    override val repository: RecipeStepRepository = mock()
//    override val logic: RecipeStepLogic = spy(DefaultRecipeStepLogic(repository))
//
//    override val entity: RecipeStep = recipeStep(id = 0L, message = "message")
//    override val anotherEntity: RecipeStep = recipeStep(id = 1L, message = "another message")
//
//    // validateGroupInformationSteps()
//
//    @Test
//    fun `validateGroupInformationSteps() calls validateSteps() with the given RecipeGroupInformation steps`() {
//        withGroupInformation {
//            logic.validateGroupInformationSteps(this)
//
//            verify(logic).validateSteps(this.steps!!)
//        }
//    }
//
//    @Test
//    fun `validateGroupInformationSteps() throws InvalidGroupStepsPositionsException when validateSteps() throws an InvalidStepsPositionsException`() {
//        withGroupInformation {
//            doAnswer { throw InvalidStepsPositionsException(setOf()) }.whenever(logic).validateSteps(this.steps!!)
//
//            assertThrows<InvalidGroupStepsPositionsException> {
//                logic.validateGroupInformationSteps(this)
//            }
//        }
//    }
//
//    // validateSteps()
//
//    @Test
//    fun `validateSteps() throws an InvalidStepsPositionsException when the position of the first step of the given groupInformation is not 1`() {
//        assertInvalidStepsPositionsException(
//          mutableSetOf(
//            recipeStep(id = 0L, position = 0),
//            recipeStep(id = 1L, position = 1),
//            recipeStep(id = 2L, position = 2),
//            recipeStep(id = 3L, position = 3)
//          ),
//          INVALID_FIRST_STEP_POSITION_ERROR_CODE
//        )
//    }
//
//    @Test
//    fun `validateSteps() throws an InvalidStepsPositionsException when steps positions are duplicated in the given groupInformation`() {
//        assertInvalidStepsPositionsException(
//          mutableSetOf(
//            recipeStep(id = 0L, position = 1),
//            recipeStep(id = 1L, position = 2),
//            recipeStep(id = 2L, position = 2),
//            recipeStep(id = 3L, position = 3)
//          ),
//          DUPLICATED_STEPS_POSITIONS_ERROR_CODE
//        )
//    }
//
//    @Test
//    fun `validateSteps() throws an InvalidStepsPositionsException when there is a gap between steps positions in the given groupInformation`() {
//        assertInvalidStepsPositionsException(
//          mutableSetOf(
//            recipeStep(id = 0L, position = 1),
//            recipeStep(id = 1L, position = 2),
//            recipeStep(id = 2L, position = 4),
//            recipeStep(id = 3L, position = 5)
//          ),
//          GAP_BETWEEN_STEPS_POSITIONS_ERROR_CODE
//        )
//    }
//
//    private fun withGroupInformation(steps: MutableSet<RecipeStep>? = null, test: RecipeGroupInformation.() -> Unit) {
//        recipeGroupInformation(
//          group = group(id = 0L),
//          steps = steps ?: mutableSetOf(
//            recipeStep(id = 0L, position = 1),
//            recipeStep(id = 1L, position = 2),
//            recipeStep(id = 2L, position = 3),
//            recipeStep(id = 3L, position = 4)
//          )
//        ) {
//            test()
//        }
//    }
//
//    private fun assertInvalidStepsPositionsException(steps: MutableSet<RecipeStep>, errorType: String) {
//        val exception = assertThrows<InvalidStepsPositionsException> {
//            logic.validateSteps(steps)
//        }
//
//        assertTrue { exception.errors.size == 1 }
//        assertTrue { exception.errors.first().type == errorType }
//    }
//}
